#ifndef TUMBLER_H
#define TUMBLER_H

#include <functional>

#define CREATE d.position += m_offset;\
               body = m_world->CreateBody(&d);\
               m_bodies.push_back(body);\
               body->CreateFixture(&shape, density)->SetFilterData(bodypartFD);

#define BODYPARTCC 2
#define MIXER 4

#include <vector>

class Tumbler : public Test
{
    std::vector<b2Body*> m_bodies;
    std::vector<b2RevoluteJoint*> m_joints;

    static const int density = 50;

    enum
    {
        LFOOT = 0,
        LCALF,
        LFEMUR,
        TORSO,
        HEAD,
        LUPPERARM,
        LFOREARM,
        LWIST,
        RFOOT,
        RCALF,
        RFEMUR,
        RUPPERARM,
        RFOREARM,
        RWRIST
    };

    b2Body * mixer;

public:
    Tumbler()
    {  
        {
            m_relax = m_rotateMinus = false;
            m_offset.Set(-0.4f,8.0f);
            projectile = 0;
        }

        b2Body* ground;
        
        {
            b2BodyDef bd;
            ground = m_world->CreateBody(&bd);

            b2EdgeShape shape;
            shape.Set(b2Vec2(-10.0f, 0.0f), b2Vec2(-10.0f, 14.0f));
            ground->CreateFixture(&shape, 0.0f);

            shape.Set(b2Vec2(10.0f, 0.0f), b2Vec2(-10.0f, 0.0f));
            ground->CreateFixture(&shape, 0.0f);

            shape.Set(b2Vec2(10.0f, 14.0f), b2Vec2(10.0f, 0.0f));
            ground->CreateFixture(&shape, 0.0f);

            shape.Set(b2Vec2(-4.0f, 7.0f), b2Vec2(0.0f, 7.0f));
            ground->CreateFixture(&shape, 0.0f);
        }

        {
            b2BodyDef bd;
            b2Body* ground = m_world->CreateBody(&bd);

            float32 x1 = -10.0f;
            float32 y1 = 2.0f * cosf(x1 / 5.0f * b2_pi) + 12.0f;
            for (int32 i = 0; i < 80; ++i)
            {
                float32 x2 = x1 + 0.25f;
                float32 y2 = 2.0f * cosf(x2 / 5.0f * b2_pi) + 12.0f;

                b2EdgeShape shape;
                shape.Set(b2Vec2(x1, y1), b2Vec2(x2, y2));
                ground->CreateFixture(&shape, 0.0f);

                x1 = x2;
                y1 = y2;
            }
        }

        {
            b2BodyDef d;
            d.type = b2_staticBody;
            b2PolygonShape shape;
            d.position.Set(10.0f,0.0f);
            b2Body * stairs = m_world->CreateBody(&d);

            for(unsigned i = 2; i < 10; ++i)
            {
                shape.SetAsBox(i / 2.0f, 5.0f - i / 2.0f);
                stairs->CreateFixture(&shape, 0.0f);
            }

            d.position.Set(-5.0f, 0.0f);
            shape.SetAsBox(2.0,2.0);
            d.angle = 0.3f;
            m_world->CreateBody(&d)->CreateFixture(&shape, 0.0f);

            d.position.Set(-2.0f, 0.0f);
            shape.SetAsBox(2.0,1.0);
            d.angle = 0.03f;
            m_world->CreateBody(&d)->CreateFixture(&shape, 0.0f);
        }

        {
            b2Filter mixerFD;
            mixerFD.categoryBits = MIXER;
            mixerFD.maskBits = BODYPARTCC;

            b2BodyDef d;
            d.type = b2_dynamicBody;
            b2PolygonShape shape;
            d.position.Set(6.0f,3.0f);
            shape.SetAsBox(1.5,0.2);
            mixer = m_world->CreateBody(&d);
            mixer->CreateFixture(&shape, 1.0f)->SetFilterData(mixerFD);

            b2RevoluteJointDef rjd;
            rjd.bodyA = ground;
            rjd.bodyB = mixer;
            rjd.localAnchorA.Set(2.0f,0.0f);
            rjd.localAnchorB.Set(0.9f, 0.0f);
            rjd.enableMotor = true;
            rjd.motorSpeed = - 3.0f * b2_pi;
            rjd.enableLimit = false;
            rjd.collideConnected = false;
            rjd.maxMotorTorque = 2000;
            m_world->CreateJoint(&rjd);
        }

        createDummy();


    }

    void createDummy() 
    {
        b2Filter bodypartFD;
        bodypartFD.categoryBits = BODYPARTCC;
        bodypartFD.maskBits = 1 | MIXER;

        {  
            //lfoot
            b2Body * body;
            b2BodyDef d;
            d.type = b2_dynamicBody;
            d.position.Set(0.0f,0.1f);
            d.position += m_offset;
            body = m_world->CreateBody(&d);
            m_bodies.push_back(body);
            b2PolygonShape shape;
            shape.SetAsBox(0.15f, 0.04f);
            body->CreateFixture(&shape, density)->SetFilterData(bodypartFD);

            //lcalf
            d.position.Set(0.07f, 0.32f);
            shape.SetAsBox(0.06f, 0.22f);
            CREATE;

            //lfemur
            d.position.Set(0.07f, 0.65f);
            shape.SetAsBox(0.07f, 0.23f);
            CREATE;

            //torso
            d.position.Set(0.07f, 1.1f);
            shape.SetAsBox(0.12f, 0.33f);
            CREATE;

            //head
            d.position.Set(0.07f, 1.7f);
            shape.SetAsBox(0.15f, 0.15f);
            CREATE;

            //lupperarm
            d.position.Set(0.07f, 1.2f);
            shape.SetAsBox(0.06f, 0.17f);
            CREATE;

            //lforearm
            d.position.Set(0.07f, 0.90f);
            shape.SetAsBox(0.05f, 0.17f);
            CREATE;

            //lwrist
            d.position.Set(0.07f, 0.70f);
            shape.SetAsBox(0.065f, 0.06f);
            CREATE;

            //rfoot
            d.position.Set(0.0f,0.1f);
            shape.SetAsBox(0.15f, 0.04f);
            CREATE;

            //rcalf
            d.position.Set(0.07f, 0.32f);
            shape.SetAsBox(0.06f, 0.22f);
            CREATE;

            //rfemur
            d.position.Set(0.07f, 0.65f);
            shape.SetAsBox(0.07f, 0.23f);
            CREATE;

            //rupperarm
            d.position.Set(0.07f, 1.2f);
            shape.SetAsBox(0.06f, 0.17f);
            CREATE;

            //rforearm
            d.position.Set(0.07f, 0.90f);
            shape.SetAsBox(0.05f, 0.17f);
            CREATE;

            //rwrist
            d.position.Set(0.07f, 0.70f);
            shape.SetAsBox(0.065f, 0.06f);
            CREATE;

        }

        {
            b2RevoluteJointDef rjd;
            rjd.enableLimit = true;
            rjd.collideConnected = false;

            rjd.bodyA = m_bodies[LFOOT];
            rjd.bodyB = m_bodies[LCALF];
            rjd.lowerAngle = -0.3f * b2_pi;
            rjd.upperAngle = 0.15f * b2_pi;
            rjd.localAnchorA.Set(0.12f, 0.03f);
            rjd.localAnchorB.Set(0.0f, -0.18f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[LCALF];
            rjd.bodyB = m_bodies[LFEMUR];
            rjd.lowerAngle = -0.9f * b2_pi;
            rjd.upperAngle = 0.01f * b2_pi;
            rjd.localAnchorA.Set(0.0f, 0.18f);
            rjd.localAnchorB.Set(0.0f, -0.22f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[LFEMUR];
            rjd.bodyB = m_bodies[TORSO];
            rjd.lowerAngle = -0.1f * b2_pi;
            rjd.upperAngle = 0.75f * b2_pi;
            rjd.localAnchorA.Set(0.0f, .22f);
            rjd.localAnchorB.Set(0.0f, -0.32f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));


            rjd.bodyA = m_bodies[RFOOT];
            rjd.bodyB = m_bodies[RCALF];
            rjd.lowerAngle = -0.3f * b2_pi;
            rjd.upperAngle = 0.15f * b2_pi;
            rjd.localAnchorA.Set(0.12f, 0.03f);
            rjd.localAnchorB.Set(0.0f, -0.18f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[RCALF];
            rjd.bodyB = m_bodies[RFEMUR];
            rjd.lowerAngle = -0.9f * b2_pi;
            rjd.upperAngle = 0.01f * b2_pi;
            rjd.localAnchorA.Set(0.0f, 0.18f);
            rjd.localAnchorB.Set(0.0f, -0.22f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[RFEMUR];
            rjd.bodyB = m_bodies[TORSO];
            rjd.lowerAngle = -0.1f * b2_pi;
            rjd.upperAngle = 0.75f * b2_pi;
            rjd.localAnchorA.Set(0.0f, .22f);
            rjd.localAnchorB.Set(0.0f, -0.32f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[TORSO];
            rjd.bodyB = m_bodies[HEAD];
            rjd.lowerAngle = -0.25f * b2_pi;
            rjd.upperAngle = 0.25f * b2_pi;
            rjd.localAnchorA.Set(0.0f, 0.32f);
            rjd.localAnchorB.Set(0.0f, -0.19f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[TORSO];
            rjd.bodyB = m_bodies[LUPPERARM];
            rjd.lowerAngle = -0.99f * b2_pi;
            rjd.upperAngle = 0.15f * b2_pi;
            rjd.localAnchorA.Set(0.0f, 0.30f);
            rjd.localAnchorB.Set(0.0f, 0.16f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[LUPPERARM];
            rjd.bodyB = m_bodies[LFOREARM];
            rjd.lowerAngle = -0.9f * b2_pi;
            rjd.upperAngle = 0.0f * b2_pi;
            rjd.localAnchorA.Set(0.0f, -0.16f);
            rjd.localAnchorB.Set(0.0f, 0.16f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[LFOREARM];
            rjd.bodyB = m_bodies[LWIST];
            rjd.lowerAngle = -0.2f * b2_pi;
            rjd.upperAngle = 0.2f * b2_pi;
            rjd.localAnchorA.Set(0.0f, -0.16f);
            rjd.localAnchorB.Set(0.0f, 0.05f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[TORSO];
            rjd.bodyB = m_bodies[RUPPERARM];
            rjd.lowerAngle = -0.99f * b2_pi;
            rjd.upperAngle = 0.15f * b2_pi;
            rjd.localAnchorA.Set(0.0f, 0.30f);
            rjd.localAnchorB.Set(0.0f, 0.16f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[RUPPERARM];
            rjd.bodyB = m_bodies[RFOREARM];
            rjd.lowerAngle = -0.9f * b2_pi;
            rjd.upperAngle = 0.0f * b2_pi;
            rjd.localAnchorA.Set(0.0f, -0.16f);
            rjd.localAnchorB.Set(0.0f, 0.16f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));

            rjd.bodyA = m_bodies[RFOREARM];
            rjd.bodyB = m_bodies[RWRIST];
            rjd.lowerAngle = -0.2f * b2_pi;
            rjd.upperAngle = 0.2f * b2_pi;
            rjd.localAnchorA.Set(0.0f, -0.16f);
            rjd.localAnchorB.Set(0.0f, 0.05f);
            m_joints.push_back((b2RevoluteJoint*)m_world->CreateJoint(&rjd));
        }
    }

    b2Body * projectile;

    void fire()
    {
        if(!projectile)
        {
            b2Filter bf;
            bf.categoryBits = MIXER;
            bf.maskBits = BODYPARTCC;
            b2BodyDef d;
            d.type = b2_dynamicBody;
            d.bullet = true;
            projectile = m_world->CreateBody(&d);
            b2CircleShape cs;
            cs.m_radius = 0.15;
            projectile->CreateFixture(&cs, 50.0f)->SetFilterData(bf);
            projectile->SetGravityScale(0.0f);
        }
        projectile->SetTransform(b2Vec2(-9.0f,5.5f), 0.0f);
        projectile->SetLinearVelocity(b2Vec2(0.0f,0.0f));
        projectile->ApplyLinearImpulse(projectile->GetMass() * b2Vec2(50.0f, 0.0f), projectile->GetWorldCenter(), true);
    }

    void clear()
    {
        //std::for_each(m_bodies.begin(), m_bodies.end(), [&](b2Body*b){m_world->DestroyBody(b);});
        m_bodies.clear();
        m_joints.clear();
    }


    static Test* Create()
    {
        return new Tumbler;
    }

    virtual void Step( Settings* settings ) 
    {
        Test::Step(settings);
        static unsigned i = 0;
        if(++i%60 == 0)
        {
            fire();
        }

        if(mixer->GetAngularVelocity() > -0.01f && mixer->GetAngularVelocity() < 0.01f)
        {
            b2RevoluteJoint * j = static_cast<b2RevoluteJoint*>(mixer->GetJointList()->joint);
            j->SetMotorSpeed(-j->GetMotorSpeed());
        }

        m_debugDraw.DrawString(5, m_textLine, "press 'e' to spawn dummy on the position of last click");
        m_textLine += 15;
        m_debugDraw.DrawString(5, m_textLine, "press 'g' to create gears");
        m_textLine += 15;
    }

    virtual void Keyboard( unsigned char key ) 
    {
        Test::Keyboard(key);
        if(key == 'e')
        {
            clear();
            createDummy();
        }

        static bool gears = false;
        if(key == 'g' && !gears)
        {
            gears = true;
            generateGear(b2Vec2(-20.0f, 5.0f), 5.0f, 6.0f, 24, true);
            generateGear(b2Vec2(-31.5f, 5.0f), 5.0f, 6.0f);

            generateGear(b2Vec2(-22.8f, 25.0f), 2.0f, 2.5f, 14, true);
            generateGear(b2Vec2(-27.5f, 25.0f), 2.0f, 2.5f, 14);
        }
    }

    virtual void KeyboardUp( unsigned char key ) 
    {
        Test::KeyboardUp(key);
        if(key == 'q')
            m_relax = false;
    }

    virtual void MouseMove(const b2Vec2& p)
    {
        Test::MouseDown(p);

        m_offset = p;
    }

    virtual void MouseDown(const b2Vec2& p)
    {
        Test::MouseDown(p);

        m_offset = p;
       
    }

    void generateGear(const b2Vec2& c, float radius, float outerRadius, unsigned tts = 24, bool motor = false)
    {
        unsigned vertices = tts * 4;
        float tangle = b2_pi * 2.0f / vertices;
        float currangle = 0.0f;

        b2Filter bf;
        bf.categoryBits = MIXER;
        bf.maskBits = BODYPARTCC | MIXER;
        b2BodyDef d;
        d.position = c;
        d.type = b2_dynamicBody;
        d.bullet = true;
        b2Body * gear = m_world->CreateBody(&d);

        for (unsigned i = 0; i < tts; ++i)
        {
            b2PolygonShape s;
            b2Vec2 vertices[4];
            vertices[0].Set(radius * cos(currangle),  radius * sin(currangle)); currangle += tangle;
            vertices[1].Set(outerRadius * cos(currangle),  outerRadius * sin(currangle)); currangle += tangle;
            vertices[2].Set(outerRadius * cos(currangle),  outerRadius * sin(currangle)); currangle += tangle;
            vertices[3].Set(radius * cos(currangle),  radius * sin(currangle)); currangle += tangle;
            s.Set(vertices, 4);
            gear->CreateFixture(&s, 50.0f)->SetFilterData(bf);
        }

        b2CircleShape cs;
        cs.m_radius = radius;
        gear->CreateFixture(&cs, 50.0f)->SetFilterData(bf);

        b2RevoluteJointDef jd;
        jd.bodyA = m_groundBody;
        jd.bodyB = gear;
        jd.localAnchorA = c;
        jd.localAnchorB.Set(0.0f,0.0f);
        jd.enableLimit = false;
        jd.collideConnected = true;
        if(motor)
        {
            jd.motorSpeed = b2_pi / 10.0f;
            jd.enableMotor = true;
            jd.maxMotorTorque = 2000000;
        }
        b2RevoluteJointDef rjd;
        m_world->CreateJoint(&jd);
    }

    bool m_relax;
    bool m_rotateMinus;
    b2Vec2 m_offset;
};

#endif
